#
# ECRTS 2010 submission "Sulphur"
# Results generator
# Copyright (C) Jack Whitham 2010
#

import pickle, math
from tetra import ASSIGNMENT, FN_LIST

CACHE_KEY = "acet1"
KWRK = 1.5
PC = 1.05
PCS = "%1.0f%%" % ((PC - 1.0) * 100.0)
HISTO_INTERVAL = 20.0
REFERENCE = 1.0
BUSLAT = 49
fn_list = FN_LIST
assignment = dict([ (x[ 0 ], x) for x in ASSIGNMENT ])
todo2 = dict()
execfile("memlab_cdata.py", { "data" : todo2 })

def Pre_Proc(name):
    if len(name) > 12:
        name = name[ : 9 ] + "..."
    name = name.replace('%', '\\%').replace('_', '\\_')
    name = name.rjust(15)
    return name

def Resolve(name):
    v = assignment[ name ]
    return ('%d,%s,%s' % (v[ 1 ], v[ 2 ], v[ 3 ]))[ :30 ].ljust(30)
   
print 'Write models table'
codes = file("sulphurdata/codes2.tex", "wt")
codes.write(r"\newcommand{\mumu}{%1.0f}" % ((PC - 1.0) * 100.0))
models = file("sulphurdata/models.tex", "wt")
models.write(
r"\begin{tabular}{|c|c|c|c|}\hline" "\n"
r"\ts Model &\ts SMMU &\ts Cache &\ts SMMU WCET / \\" "\n"
r"\ts Nr    &\ts WCET &\ts ACET &\ts Cache ACET \\" "\n"
r"\hline" "\n")
models2 = file("sulphurdata/models2.tex", "wt")
models2.write(
r"\begin{tabular}{|c|c|c|c|}\hline" "\n"
r"\ts Model &\ts Tiled &\ts Cache   & \ts SMMU WCET / \\" "\n"
r"\ts Nr    &\ts SMMU WCET &\ts ACET& \ts Cache ACET \\" "\n"
r"\hline" "\n")
models3 = file("sulphurdata/models3.tex", "wt")
models3.write(
r"\begin{tabular}{|c|c|c|c|}\hline" "\n"
r"\ts Model &\ts Modified &\ts Cache &  SMMU / \\" "\n"
r"\ts Nr    &\ts SMMU ACET  &\ts ACET  &  Cache \\" "\n"
r"\hline" "\n")
options = file("sulphurdata/options.tex", "wt")
options.write(
r"\begin{tabular}{|c|c|c|c|}\hline" "\n"
r"\ts Model &\ts Option &\ts Option     &\ts Cache \\" "\n"
r"\ts Nr    &\ts (b) WCET &\ts (c) WCET &\ts ACET \\" "\n"
r"\hline" "\n")
final = file("sulphurdata/final.tex", "wt")
final.write(
r"\begin{tabular}{|c|c|c|c|}\hline" "\n"
r"\ts Model &\ts SMMU WCET / &\ts Model &\ts SMMU WCET / \\" "\n"
r"\ts Nr    &\ts Cache WCET  &\ts Nr    &\ts Cache WCET \\" "\n"
r"\hline" "\n")
csvs = [ file("sulphurdata/csv%d.csv" % i, "wt") for i in xrange(6) ]
desc = file("sulphurdata/summary1.tex", "wt")
desc.write(
r"\begin{tabular}{|c|ll|p{3.2cm}|}\hline" "\n"
r"\ts Nr &\ts  Program &\ts  Function &\ts  Contents \\" "\n"
r"\hline" "\n")

fn_list_2 = fn_list[:]
fn_list_2.sort()
fn_list_2.sort(key = lambda test_fn_name: 
        - (todo2[ (test_fn_name, "Cache_16384", BUSLAT) ] /
        todo2[ (test_fn_name, "SMMU_F", BUSLAT) ]))
worse_count = 0
tiled_ok_count = tiled_great_count = 0
dyn_ok_count = better_count = 0
seen_before_set = set()
rnd_numbers = []
tiled_improv = dyn_improv = 0
b_numbers = []
close_together = []
cache_smmu_comparison = []
options_b = []
options_c = []
woot = 0
number = 0
for test_fn_name in fn_list_2:
    try:
        smmu = todo2[ (test_fn_name, "SMMU_F", BUSLAT) ]
        cache = float(todo2[ (test_fn_name, "Cache_16384", BUSLAT) ])
        smmu_tiled = todo2[ (test_fn_name, "SMMU_TF", BUSLAT) ]
        smmu_b = todo2[ (test_fn_name, "SMMU_16384", BUSLAT) ]
    except TypeError, KeyError:
        continue

    (ncode, _, program, func_name, code) = assignment[ test_fn_name ]
    unique = ','.join(code)

    if unique in seen_before_set:
        print 'Reject', ncode
        continue

    seen_before_set.add(unique)

    identifiers = assignment[ test_fn_name ][ -1 ]
    identifiers = identifiers[:]
    number += 1
    v = assignment[ test_fn_name ]
    notes = []
    notes2 = []
    notes3 = []

    if func_name == "susan_smoothing":
        codes.write(r'\newcommand{\asbo}{%d}' % number)

    margin = cache * PC
    if smmu > margin:
        worse_count += 1
    else:   
        notes.append("OK")

    if smmu < (cache * 1.005):
        close_together.append(str(number))
        close_together.append(", ")

    if smmu_tiled <= margin:
        tiled_ok_count += 1
        notes2.append("OK")

    if smmu_b <= margin:
        dyn_ok_count += 1
        notes3.append("OK")

    if (smmu / smmu_tiled) >= PC:
        tiled_improv += 1
    if (smmu_tiled / smmu_b) >= PC:
        dyn_improv += 1

    #if (smmu_tiled * 1.1) <= cache:
    #    notes2.append("E")
    #    print 'boo'
    #    tiled_great_count += 1

    # How to generate NF flag? Does it matter?

    seq_access = False
    rd_access = False
    d_access = False
    r_access = False
    for ident in identifiers:
        if 'S' in ident:
            seq_access = True
        if ('R' in ident) or ('D' in ident):
            rd_access = True
        if ('R' in ident):
            r_access = True
    if seq_access:
        notes.append("SA")
    if smmu_tiled > margin:
        if r_access:
            rnd_numbers.append('%d' % number)
            rnd_numbers.append(', ')
        #if not rd_access:
        #    print 'Exception on number %d' % number
    if smmu_b >= cache:
        better_count += 1

    b_numbers.append(smmu_b / cache)

    #identifiers.sort()
    #print number, ' '.join(identifiers)

    if smmu_tiled <= cache:
        woot += 1

    m1 = (r"\ts %d &\ts %1.2f &\ts %1.2f &\ts %1.2f \\" "\n" % (
            number, smmu, cache, 1.0 / (smmu / cache)))
    m2 = (r"\ts %d &\ts %1.2f &\ts %1.2f &\ts %1.2f \\" "\n" % (
            number, smmu_tiled, cache, 1.0 / (smmu_tiled / cache)))
    m3 = (r"\ts %d &\ts %1.2f &\ts %1.2f &\ts %1.2f \\" "\n" % (
            number, smmu_b, cache, 1.0 / (smmu_b / cache)))

    for (csv, value) in zip(csvs[ :3 ], 
                    (smmu, smmu_tiled, smmu_b)):
        csv.write('%d,%f\n' % (
                number, cache / value))

    csvs[ 5 ].write('%d,%f\n' % (number, cache))
    models.write(m1)
    models2.write(m2)
    models3.write(m3)
    options.write(r"\ts %d &\ts %1.2f &\ts %1.2f &\ts %1.2f \\" "\n" % (
            number, smmu_b, 0, cache))
    final.write(r"\ts %d &\ts %1.2f " % (
            number, smmu_b / cache))
    if number % 2:
        final.write(r" &")
    else:
        final.write(r"\\" "\n")
    cache_smmu_comparison.append((smmu_b / cache, number))

    (ncode, _, program, func_name, code) = assignment[ test_fn_name ]
    desc.write(r'\ts %d &\ts %s &\ts %s &\ts %s \\' % (number,
                program, Pre_Proc(func_name), ' '.join(code)))
    desc.write('\n')

    options_b.append(smmu_b / cache)
    options_c.append(0/ cache) # XXX

print ''
print ''
print ''

for and_list in (rnd_numbers, close_together):
    print 'and list', len(and_list)
    and_list.pop()
    last_number = and_list.pop()
    and_list.pop()
    and_list.append(" and ")
    and_list.append(last_number)

#assert len(dyn_b_different) == 1, dyn_b_different
#assert len(dyn_b_different) == 0

cache_smmu_comparison.sort()
histo = dict()
fahren = 0
for (x, number) in cache_smmu_comparison:
    histo_group = int(math.floor((x + 0.025) * HISTO_INTERVAL))
    print '%3d  %1.3f   g %u' % (number, x, histo_group)
    histo[ histo_group ] = histo.get(histo_group, 0) + 1
    if x > KWRK:
        codes.write(r"\newcommand{\kwrk}{%u}" % number)
        codes.write(r"\newcommand{\bahn}{%1.2f}" % x)
    if x < 1.0:
        fahren += 1

codes.write(r"\newcommand{\fahr}{%u}" % fahren)

histo_out = file("sulphurdata/histo.csv", "wt")
histo_keys = histo.keys()
histo_keys.sort()
for key in xrange(min(histo_keys), max(histo_keys) + 1, 1):
    value = histo.get(key, 0)
    histo_out.write("%1.2f,%d\n" % (key / HISTO_INTERVAL, value))
histo_out.close()

assert woot >= 4, woot # Hard-coded in document
codes.write(r"\newcommand{\ybpq}{%u}" % worse_count)
codes.write(r"\newcommand{\ypbq}{\ybpq}")
codes.write(r"\newcommand{\txjt}{%u}" % tiled_ok_count)
#codes.write(r"\newcommand{\qveu}{%u}" % tiled_great_count)
codes.write(r"\newcommand{\nzie}{%u}" % dyn_ok_count)
codes.write(r"\newcommand{\xyza}{%u}" % better_count)
#codes.write(r"\newcommand{\fzzl}{%u}" % dyn_b_different[ 0 ])
codes.write(r"\newcommand{\sjdf}{%s}" % ''.join(rnd_numbers))
codes.write(r"\newcommand{\closetogether}{%s}" % ''.join(close_together))
codes.write(r"\newcommand{\woot}{%u}" % woot)
codes.write(r"\newcommand{\nummodels}{%u}" % len(options_b))
codes.write(r"\newcommand{\xooy}{%u}" % (tiled_improv))
codes.write(r"\newcommand{\xooz}{%u}" % (dyn_improv))
codes.write(r"\newcommand{\smub}{%1.2f}" % (
        (sum(b_numbers) / len(b_numbers))))

options_b.sort()
options_c.sort()
assert len(options_b) == len(options_c)

for i in [ 0.0, 0.25, 0.5, 0.75, 0.9, 1.0 ]:
    pos = int(math.floor(i * (len(options_b) - 1)))
    assert 0 <= pos < len(options_b)
    csvs[ 3 ].write('%1.0fth percentile,' % (i * 100.0))
    csvs[ 3 ].write('%f,%f\n' % (
            options_b[ pos ], options_c[ pos ]))
csvs[ 3 ].write('Mean,%f,%f\n' % (
        sum(options_b) / len(options_b),
        sum(options_c) / len(options_c)))

for out in (models, models2, models3, options, final, desc):
    out.write(r"\hline\end{tabular}" "\n\n")
    out.close()
codes.close()

overall_csv = file("sulphurdata/overall.csv", "wt")

fig6 = file("sulphurdata/fig6.dat", "wt")

#overall_csv.write("Size,10th,mean,90th\n")

buslat = BUSLAT
smmu_best = dict()
all_sizes = [ 1 << x for x in [ 9, 10, 11, 12, 13, 14, 15, 16, 17] ]
for size in all_sizes:
    #overall_csv.write("%d" % size)
    smmu_key = "SMMU_%d" % size
    cache_key = "Cache_%d" % size

    data = []
    for test_fn_name in fn_list:
        try:
            smmu_val = todo2[ (test_fn_name, smmu_key, buslat) ]
            cache_val = float(todo2[ (test_fn_name, cache_key, buslat) ])
        except:
            continue

        data.append(cache_val / smmu_val)

    assert len(data) != 0, "no data for size %u" % size
    data.sort()
    mean = sum(data) / len(data)
    sd = math.sqrt(sum([ (d - mean) ** 2 for d in data ]) / len(data))
    #ci = sd * 1.96 # 95% confidence
    ten_percent = int(math.ceil(len(data) * 0.1))
    print '10%% actually is %1.2f%%' % (
            (ten_percent * 100.0) / len(data))
    data = data[ : -ten_percent ]
    data = data[ ten_percent : ]
    fig6.write('%d %f %f %f\n' % (size,
                mean,
                data[ 0 ],
                data[ -1 ]))

overall_csv.write("Buslat,10th,mean,90th\n")
out_mean = []
out_sd = []
out_med = []

for buslat in [BUSLAT / 2, BUSLAT, BUSLAT * 2]:
    overall_csv.write("%d" % buslat)
    smmu_key = "SMMU_16384"
    cache_key = "Cache_16384"
    for test_fn_name in fn_list:
        try:
            smmu_val = todo2[ (test_fn_name, smmu_key, buslat) ]
            cache_val = float(todo2[ (test_fn_name, cache_key, buslat) ])
        except:
            continue

        data.append(cache_val / smmu_val)

    data.sort()
    mean = sum(data) / len(data)
    sd = math.sqrt(sum([ (d - mean) ** 2 for d in data ]) / len(data))

    out_mean.append(mean)
    out_sd.append(sd)
    out_med.append(data[ len(data) / 2 ])

for i in xrange(1, len(out_mean), 1):
    print '%f' % (out_mean[ i ] / out_mean[ i - 1 ]),
print ''
for i in xrange(1, len(out_sd), 1):
    print '%f' % (out_sd[ i ] / out_sd[ i - 1 ]),
print ''
for i in xrange(1, len(out_med), 1):
    print '%f' % (out_med[ i ] / out_med[ i - 1 ]),
print ''

for buslat in [BUSLAT / 2, BUSLAT, BUSLAT * 2]:
    overall_csv.write("\nBuslat,%d\n" % buslat)
    for size in all_sizes:
        smmu_key = "SMMU_%d" % size
        cache_key = "Cache_%d" % size

        overall_csv.write("%d,S" % size)
        for test_fn_name in fn_list:
            try:
                smmu_val = todo2[ (test_fn_name, smmu_key, buslat) ]
            except:
                continue
            overall_csv.write(",%f" % smmu_val)

        overall_csv.write("\n%d,C" % size)
        for test_fn_name in fn_list:
            try:
                cache_val = todo2[ (test_fn_name, cache_key, buslat) ]
            except:
                continue
            overall_csv.write(",%f" % cache_val)

        overall_csv.write("\n")

    overall_csv.write("\n")



